/*******************************************************************************
 * Copyright 2024 ModalAI Inc.
 ******************************************************************************/

/**
 * @file        mcv-test-feature-tracker.cpp
 *
 * @brief       Test code for the modalcv feature tracker
 *
 * @author      Thomas Patton (thomas.patton@modalai.com)
 * @date        2023
 */

#include <fcntl.h>
#include <getopt.h>
#include <modalcv.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include <iomanip>
#include <iostream>
#include <modalcv/io.hpp>
#include <opencv2/imgcodecs.hpp>
#include <opencv2/imgproc.hpp>
#include <sstream>

#include "modalcv/feature_tracker.h"

#define OUTPUT_FOLDER "/data/libmodalcv/feature_tracker/"
#define N_DISPLAY 50  // number of fp to draw

static int64_t _apps_time_monotonic_ns() {
    struct timespec ts;
    if (clock_gettime(CLOCK_MONOTONIC, &ts)) {
        fprintf(stderr, "ERROR calling clock_gettime\n");
        return -1;
    }
    return (int64_t)ts.tv_sec * 1000000000 + (int64_t)ts.tv_nsec;
}

/**
 * @brief Draws outputted features onto a color image
 *
 * @param img Image to draw on
 * @param output_features The output features to draw
 * @param img Number of outputted features
 */
static void draw_features(
    cv::Mat img,
    mcv_cvp_feature_tracker_output_t output_features[MAX_N_FEATURES]) {
    for (int16_t i = 0; i < N_DISPLAY; i++) {
        mcv_cvp_feature_tracker_output_t feature = output_features[i];
        int age = feature.age;
        int radius, thickness;
        if (age > 15) {
            radius = 4;
            thickness = 2;
        } else {
            radius = 1;
            thickness = 1;
        }
        cv::Scalar color(std::max(0, 255 - 5 * age), 255, 0);
        cv::Rect rect(feature.x - radius, feature.y - radius, radius * 2,
                      radius * 2);
        cv::rectangle(img, rect, color, thickness, 1, 0);
    }
}

int main(int argc, char *argv[]) {
    mcv_en_debug_prints(1);
    mcv_en_timing_prints(1);

    // vars to hold args
    int w, h;
    std::vector<std::string> paths;

    // parse input arguments and extract paths for feature extraction
    if (init_modalcv_example(argc, argv, paths, w, h)) {
        exit(1);
    }

    // configuration object for feature tracking
    mcv_cvp_feature_tracker_config_t feature_tracker_config;

    // which octave of the pyramid to perform optic flow on
    feature_tracker_config.opt_flow_octave = 2;

    // create a binary mask for features (NULL implies all 1's)
    cv::Mat mask = cv::Mat::ones(h, w, CV_8UC1) * 255;
    cv::Rect lower_right_corner(w / 2, h / 2, w / 2, h / 2);
    mask(lower_right_corner).setTo(cv::Scalar(0));
    feature_tracker_config.feature_mask = mask.data;

    // set up feature point extraction
    feature_tracker_config.pyr_fpx_config.width = w;
    feature_tracker_config.pyr_fpx_config.height = h;
    feature_tracker_config.pyr_fpx_config.n_octaves = 3;
    feature_tracker_config.pyr_fpx_config.n_scales_per_octave = 1;
    feature_tracker_config.pyr_fpx_config.mode = MCV_CVP_FPX_ZONE;
    feature_tracker_config.pyr_fpx_config.nms_mode = MCV_CVP_FPX_5_TAP_NMS;
    feature_tracker_config.pyr_fpx_config.score_threshold = 1;
    feature_tracker_config.pyr_fpx_config.robustness = 20;
    feature_tracker_config.pyr_fpx_config.output_ubwc = 1;
    feature_tracker_config.pyr_fpx_config.zone_dims[0] = 16;
    feature_tracker_config.pyr_fpx_config.zone_dims[1] = 8;
    feature_tracker_config.pyr_fpx_config.zone_dims[2] = 4;
    memset(feature_tracker_config.pyr_fpx_config.n_filter_coeffs, 3,
           MCV_CVP_MAX_PYRAMID_OCTAVES * sizeof(uint32_t));
    memset(feature_tracker_config.pyr_fpx_config.enable_fpx, 1,
           MCV_CVP_MAX_PYRAMID_OCTAVES);

    // set up opt flow
    // w/h here are dictated by the octave we want to do flow on as defined 
    // above in the feature_tracker_config
    feature_tracker_config.opt_flow_config.width =
        w / pow(2, feature_tracker_config.opt_flow_octave);
    feature_tracker_config.opt_flow_config.height =
        h / pow(2, feature_tracker_config.opt_flow_octave);
    feature_tracker_config.opt_flow_config.en_stats = 0;
    feature_tracker_config.opt_flow_config.input_ubwc =
        1;  // pyrfpx output ubwc

    // get a handle for our patch tracker
    mcv_cvp_feature_tracker_handle handle;
    handle = mcv_cvp_feature_tracker_init(feature_tracker_config);
    if (NULL == handle) {
        fprintf(stderr, "ERROR getting patch tracker handle\n");
        return -1;
    }

    //  iterate over all images in folder
    printf("Starting main loop...\n");
    for (size_t i = 0; i < paths.size(); i++) {
        // load the file
        cv::Mat current_img_gray, current_img_color;
        current_img_gray = imread(paths[i], cv::IMREAD_GRAYSCALE);
        if (current_img_gray.type() != CV_8UC1) {
            fprintf(stderr, "ERROR loading image\n");
            return -1;
        }
        cv::cvtColor(current_img_gray, current_img_color, cv::COLOR_GRAY2BGR);
        // cv::blur(current_img_gray, current_img_gray, cv::Size(5, 5));

        // some output things
        int16_t num_outputs;
        mcv_cvp_feature_tracker_output_t output[MAX_N_FEATURES];

        // run the patch tracker process on the current frame
        int64_t start = _apps_time_monotonic_ns();
        mcv_cvp_feature_tracker_process(handle, current_img_gray.data,
                                        &num_outputs, output);
        int64_t end = _apps_time_monotonic_ns();
        printf("Iter Time ::  %6.2fms\n", (double)(end - start) / 1000000);

        // draw features on the image
        draw_features(current_img_color, output);

        std::stringstream ss;
        ss << std::setw(5) << std::setfill('0') << i;
        std::string output_path = OUTPUT_FOLDER + ss.str() + ".png";
        cv::imwrite(output_path, current_img_color);
    }

    // cleanup
    printf("all done, deinitializing\n");
    mcv_cvp_feature_tracker_deinit(handle);
    return 0;
}
